/*****************************************************************************
 *            Created: 8/05/2012
 *          Copyright: (C) Copyright David Goodwin, 2012
 *            License: GNU General Public License
 *****************************************************************************
 *
 *   This is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This software is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this software; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ****************************************************************************/

/* Functions for dealing with grids */

#include "grid.h"

#include <QDebug>
#include <QFile>
#include <QMessageBox>

typedef unsigned char byte_t;


grid_t make_grid(unsigned int width, unsigned int height) {
    grid_t grid;

    qDebug() << "Creating" << width << "x" << height << "grid";

    grid = (grid_t)calloc(width,sizeof(bool*));
    for (unsigned int x = 0; x < width; x++)
        grid[x] = (bool*)calloc(height,sizeof(bool));

    return grid;
}

void free_grid(grid_t grid, unsigned int width) {
    for (unsigned int x = 0; x < width; x++)
        free((void*)grid[x]);
    free(grid);
}

void copy_grid(grid_t target, grid_t source, unsigned int width, unsigned int height) {
    for (unsigned int x = 0; x < width; x++)
        memcpy(target[x],source[x],sizeof(bool) * height);
}

grid_t duplicate_grid(grid_t grid, unsigned int width, unsigned int height) {
    grid_t new_grid = make_grid(width,height);

    copy_grid(new_grid, grid, width, height);

    return new_grid;
}

grid_t resize_grid(grid_t grid, unsigned int old_width, unsigned int old_height,
                   unsigned int new_width, unsigned int new_height) {
    grid_t new_grid = make_grid(new_width, new_height);

    for (unsigned int x = 0; x < old_width && x < new_width; x++)
        for (unsigned int y = 0; y < old_height && y < new_height; y++)
            new_grid[x][y] = grid[x][y];

    return new_grid;
}

byte_t set_bit(byte_t byte, int bit) {
    if (bit > 7 || bit < 0)
        return byte; // Invalid bit.

    return byte | 1 << bit;
}

bool get_bit(byte_t byte, int bit) {
    if (bit > 7 || bit < 0)
        return false; // Invalid bit.

    int i = byte >> (bit) & 0x00000001;

    return (i == 1);
}

/* Writes the grid to a v1 file.
 * Grid file format is:
 *   +------------+--------+--------+----------------...
 *   | ZXLGF1     | width  | height | grid bitmap
 *   +------------+--------+--------+----------------...
 *    6 bytes       4 bytes  4 bytes  width*height bits
 */
void save_grid_v1(grid_t grid, uint32_t width, uint32_t height, QString filename) {
    QFile file(filename);

    if (!file.open(QFile::WriteOnly)) {
        QMessageBox::warning(NULL,"Error","Unable to open file for writing");
        return;
    }

    file.write(GRIDFILE_HDR1,6);

    file.write((const char*)&width,4);
    file.write((const char*)&height,4);

    byte_t byte = 0;
    int current_bit = -1;
    int byte_number = 0;
    for (unsigned int y = 0; y < height; y++) {
        for (unsigned int x = 0; x < width; x++) {
            if (current_bit == 7) {
                file.write((const char*)&byte,1);
                byte_number++;
                byte = 0;
                current_bit = -1;
            }
            current_bit++;

            qDebug() << "store" << x << ":" << y << "in byte" << byte_number << "bit" << current_bit;
            if (grid[x][y])
                byte = set_bit(byte, current_bit);
        }
    }
    file.write((const char*)&byte,1);
}

/* Writes the grid to a v2 file.
 * Grid file format is:
 *   +------------+--------+------...----+--------+--------+----------------...
 *   | ZXLGF2     | dsc_len| description | width  | height | grid bitmap
 *   +------------+--------+------...----+--------+--------+----------------...
 *    6 bytes      4 bytes  dsc_len bytes 4 bytes  4 bytes  width*height bits
 */
void save_grid(grid_t grid, uint32_t width, uint32_t height, QString filename,
               QString description) {
    QFile file(filename);

    if (!file.open(QFile::WriteOnly)) {
        QMessageBox::warning(NULL,"Error","Unable to open file for writing");
        return;
    }

    file.write(GRIDFILE_HDR2,6);

    QByteArray description_data = description.toAscii();
    uint32_t description_length = description_data.length();
    file.write((const char*)&description_length,4);
    if (description_length > 0)
        file.write(description_data);

    file.write((const char*)&width,4);
    file.write((const char*)&height,4);

    byte_t byte = 0;
    int current_bit = -1;
    int byte_number = 0;
    for (unsigned int y = 0; y < height; y++) {
        for (unsigned int x = 0; x < width; x++) {
            if (current_bit == 7) {
                file.write((const char*)&byte,1);
                byte_number++;
                byte = 0;
                current_bit = -1;
            }
            current_bit++;

            qDebug() << "store" << x << ":" << y << "in byte" << byte_number << "bit" << current_bit;
            if (grid[x][y])
                byte = set_bit(byte, current_bit);
        }
    }
    file.write((const char*)&byte,1);
}

/* Loads old v1 format files */
grid_t load_grid_v1(unsigned int *width, unsigned int *height, QString filename) {
    QFile file(filename);

    if (!file.open(QFile::ReadOnly)) {
        QMessageBox::warning(NULL, "Error", "Unable to open file for reading");
        return NULL;
    }

    char sig[7];
    file.read(sig,6);
    sig[6] = 0;

    uint32_t fwidth, fheight;
    file.read((char*)&fwidth,4);
    file.read((char*)&fheight,4);

    qDebug() << sig;

    if (QString(sig) != GRIDFILE_HDR1) {
        QMessageBox::warning(NULL, "Error", "Invalid grid file");
        return NULL;
    }

    grid_t grid = make_grid(fwidth, fheight);

    uint32_t cell_count = fwidth * fheight;
    byte_t byte = 0;
    int current_bit = 8;
    unsigned int x = -1, y = 0;
    int byte_number = -1;

    qDebug() << "Loading" << cell_count << "cells from file";
    for (uint32_t i = 0; i < cell_count; i++) {
        if (current_bit > 7) {
            file.read((char*)&byte,1);
            current_bit = 0;
            byte_number++;
        }

        x++;

        if (x == fwidth) {
            x = 0;
            y++;
        }

        grid[x][y] = get_bit(byte,current_bit);

        current_bit++;
    }

    *width = fwidth;
    *height = fheight;
    return grid;
}


grid_t load_grid(unsigned int *width, unsigned int *height, QString filename,
                 QString *description) {
    QFile file(filename);

    if (!file.open(QFile::ReadOnly)) {
        QMessageBox::warning(NULL, "Error", "Unable to open file for reading");
        return NULL;
    }

    char sig[7];
    file.read(sig,6);
    sig[6] = 0;

    // Handle other format versions.
    qDebug() << sig;
    QString hdr = QString(sig);
    if (hdr != GRIDFILE_HDR2) {
        if (hdr == GRIDFILE_HDR1) {
            if (description != NULL)
                *description = "";
            return load_grid_v1(width, height, filename);
        } else {
            QMessageBox::warning(NULL, "Error", "Invalid grid file");
            return NULL;
        }
    }

    uint32_t description_length;
    file.read((char*)&description_length, 4);
    if (description_length > 0) {
        if (description != NULL) {
            char* description_data = (char*)malloc(sizeof(char) * description_length+1);
            file.read(description_data, description_length);
            description_data[description_length] = 0;
            *description = QString(description_data);
            free(description_data);
        } else {
            file.seek(file.pos() + description_length);
        }
    }

    uint32_t fwidth, fheight;
    file.read((char*)&fwidth,4);
    file.read((char*)&fheight,4);

    grid_t grid = make_grid(fwidth, fheight);

    uint32_t cell_count = fwidth * fheight;
    byte_t byte = 0;
    int current_bit = 8;
    unsigned int x = -1, y = 0;
    int byte_number = -1;

    qDebug() << "Loading" << cell_count << "cells from file";
    for (uint32_t i = 0; i < cell_count; i++) {
        if (current_bit > 7) {
            file.read((char*)&byte,1);
            current_bit = 0;
            byte_number++;
        }

        x++;

        if (x == fwidth) {
            x = 0;
            y++;
        }

        grid[x][y] = get_bit(byte,current_bit);

        current_bit++;
    }

    *width = fwidth;
    *height = fheight;
    return grid;
}

grid_t crop_grid(grid_t grid,
                 unsigned int width,
                 unsigned int height,
                 unsigned int *new_width,
                 unsigned int *new_height) {

    uint32_t top = height-1,
            left = width-1,
            right = 0,
            bottom = 0;

    // Find the block closest to y = 0
    for (unsigned int y = 0; y < height; y++)
        for (unsigned int x = 0; x < width; x++)
            if (grid[x][y]) {
                if (x <= left)
                    left = x;
                if (x >= right)
                    right = x;
                if (y <= top)
                    top = y;
                if (y >= bottom)
                    bottom = y;
            }

    *new_width = right - left+1;
    *new_height = bottom - top+1;

    grid_t new_grid = make_grid(*new_width, *new_height);

    for (unsigned int y = 0; y < *new_height; y++)
        for (unsigned int x = 0; x < *new_width; x++)
            new_grid[x][y] = grid[x+left][y+top];

    return new_grid;
}


// Not ready.
//grid_t make_grid_from_life105(QStringList grid_data,
//                              uint32_t *width,     /* out */
//                              uint32_t *height,    /* out */
//                              QString *description) {  /* out */
//    QString cur = grid_data.takeFirst().trimmed();
//    if (cur != "#Life 1.05") {
//        // Invalid data.
//        *width = 0;
//        *height = 0;
//        *description = "";
//        return 0;
//    }

//    QString rule = "";

//    QString line;
//    while (!grid_data.isEmpty()) {
//        line = grid_data.takeFirst().trimmed();
//        if (line.startsWith("#D")) {
//            line = line.left(2).trimmed();
//            *description->append(line + "\n");
//        } else if (line.startsWith("#N")) {
//            rule = "23/3";
//        } else if (line.startsWith("#R")) {
//            rule = line.left(2).trimmed();
//        }
//    }


//}

//grid_t make_grid_from_text(QString grid_data,
//                           uint32_t *width,     /* out */
//                           uint32_t *height,    /* out */
//                           QString *description) {  /* out */
//    QStringList lines = grid_data.split("\b", QString::SkipEmptyParts);

//    QString cur = lines.first();
//    if (cur == "#Life 1.05")
//        return make_grid_from_life105(lines, width, height, description);
//    else if (cur == "#Life 1.06") {
//        QMessageBox::warning(0, "Unsupported", "Life 1.06 format is not supported at this time.");
//        return 0;
//        *width=0;
//        *height=0;
//        *description = "";
//    }

//    QMessageBox::warning(0, "Invalid data", "The clipboard data is not in a known format");
//    *width=0;
//    *height=0;
//    *description = "";
//    return 0;
//}
